local mod = LibStub("AceAddon-3.0"):NewAddon("MarySue", "AceComm-3.0", "AceConsole-3.0", "AceEvent-3.0", "AceHook-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("MarySue")
local ACD3 = LibStub("AceConfigDialog-3.0")

local strsplit = _G.strsplit
local strtrim = _G.strtrim
local tinsert = _G.tinsert
local tconcat = _G.table.concat
local UnitName = _G.UnitName
local IsInInstance = _G.IsInInstance

local REQUEST = "REQ"
local PREFIX = "MARYSUE"
local THROTTLE_QUERY_TIME = 300
local CLIENT_THROTTLE_QUERY_TIME = 20

local bit_band = _G.bit.band
local bit_bor = _G.bit.bor
local bit_bxor = _G.bit.bxor

local db, hive
local queryTimes = {}
local display

local RP_LEVEL = {
	[1] = L["None"],
	[2] = L["Beginner"],
	[4] = L["Casual"],
	[8] = L["Full-time"],
	[16] = L["Mature (18+)"]
}

local CHARACTER_FLAGS = {
	[32] = L["Not roleplaying"],
	[64] = L["Out of Character"],
	[128] = L["In Character"],
	[256] = L["Looking for Contact"],
	[512] = L["Storyteller"]
}

local BEHAVIOR_VALUES = {
	[1] = L["Mouseover"],
	[2] = L["Target"]
}

local function getFullName(_, who)
	local t = who == nil and db.character or hive[who]
	if not t then
		return who
	end
	who = type(who) == "string" and who or UnitName("player")
	return strtrim(("%s %s %s"):format(t.prefix or "", who, t.suffix or ""))
end

local getStatusString
do
	local level_buffer = {}
	local role_buffer = {}
	function getStatusString(flags)
		wipe(level_buffer)
		wipe(role_buffer)
		for k, v in pairs(RP_LEVEL) do
			if bit_band(flags, k) == k then
				tinsert(level_buffer, v)
			end
		end

		for k, v in pairs(CHARACTER_FLAGS) do
			if bit_band(flags, k) == k then
				tinsert(role_buffer, v)
			end
		end
		
		local t1 = tconcat(level_buffer, ", ")
		local t2 = tconcat(role_buffer, ", ")
		wipe(role_buffer)
		if t1 and #t1 > 0 then tinsert(role_buffer, t1) end
		if t2 and #t2 > 0 then tinsert(role_buffer, t2) end
		local c = tconcat(role_buffer, " :: ")
		return c == "" and L["(No status)"] or c
	end
end

local options = {
	type = "group",
	childGroups = "tab",
	args = {
		char = {
			type = "group",
			name = L["Character Profile"],
			args = {
				preview = {
					type = "header",
					name = getFullName,
					order = 1
				},
				prefix = {
					type = "input",
					name = L["Prefix"],
					desc = L["A prefix that shows before your name, like \"Lady\" or \"Crazy\""],
					get = function()
						return db.character.prefix
					end,
					set = function(info, v)
						db.character.prefix = v
					end,
					order = 10
				},
				suffix = {
					type = "input",
					name = L["Suffix"],
					desc = L["A suffix that shows after your name, like \"of Stormwind\" or \"Jinklestein\""],
					get = function()
						return db.character.suffix
					end,
					set = function(info, v)
						db.character.suffix = v
					end,
					order = 11
				},
				title = {
					type = "input",
					name = L["Title"],
					desc = L["Your formal title"],
					get = function()
						return db.character.title
					end,
					set = function(info, v)
						db.character.title = v
					end,
					order = 12,
					width = "full"
				},
				profile = {
					type = "input",
					name = L["Physical description"],
					desc = L["Describe your character's physical appearance, as others would perceive them."],
					multiline = true,
					get = function()
						return db.character.profile
					end,
					set = function(info, v)
						db.character.profile = v
					end,
					width = "full"
				}
			}
		},
		flags = {
			type = "group",
			name = L["Flags & Status"],
			args = {
				level = {
					type = "multiselect",
					name = L["Roleplaying level"],
					get = function(info, key)
						return bit_band(db.rp_level_mask, key) == key
					end,
					set = function(info, key, v)
						db.rp_level_mask = bit_bxor(db.rp_level_mask, key)
					end,
					values = RP_LEVEL
				},
				role = {
					type = "multiselect",
					name = L["Current role(s)"],
					get = function(info, key)
						return bit_band(db.role_mask, key) == key
					end,
					set = function(info, key, v)
						db.role_mask = bit_bxor(db.role_mask, key)
					end,
					values = CHARACTER_FLAGS
				}
			}
		},
		behavior = {
			type = "group",
			name = L["Behavior"],
			order = 105,
			args = {
				showOn = {
					type = "multiselect",
					name = L["Show on..."],
					get = function(info, key)
						return bit_band(db.behavior.showOn, key) == key
					end,
					set = function(info, key, val)
						db.behavior.showOn = bit_bxor(db.behavior.showOn, key)
					end,
					values = BEHAVIOR_VALUES
				},
				disableInInstances = {
					type = "toggle",
					name = L["Disable in instances"],
					get = function()
						return db.behavior.disableInInstances
					end,
					set = function(info, v)
						db.behavior.disableInInstances = v
					end
				},
				sticky = {
					type = "toggle",
					name = L["Sticky Window"],
					desc = L["Keeps the description window open until you close it."],
					get = function()
						return db.sticky
					end,
					set = function(info, v)
						db.sticky = v
						mod:UpdateWindow()
					end
				}				
			}
		}
	}
}

local defaults = {
	char = {
		character = {
			suffix = "",
			prefix = "",
			title = "",
			profile = ""
		},
		hive = {},
		sticky = false,
		rp_level_mask = 0x0,
		role_mask = 0x0,
		behavior = {
			showOn = 3,
			disableInInstances = true
		}
	}
}

function mod:OnInitialize()
	self:RegisterComm(PREFIX)
	self.db = LibStub("AceDB-3.0"):New("MarySueDB", defaults)
	db = self.db.char
	local realm = GetRealmName()
	db.hive[realm] = db.hive[realm] or {}
	hive = db.hive[realm]
	
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("MarySue", options)
	local optFrame = ACD3:AddToBlizOptions("MarySue")
end

function mod:OnEnable()
	self:HookScript(GameTooltip, "OnTooltipSetUnit")
    self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:UpdateWindow()
end

------------------------------------------------------------------------
--[[ These three hooks do the majority of the work ]]--
------------------------------------------------------------------------

function mod:OnTooltipSetUnit(frame)
	if not self:Active() or not self:UPDATE_MOUSEOVER_UNIT() then return end
	
	local t = _G[frame:GetName() .. "TextLeft1"]
	local name = getFullName(nil, UnitName("mouseover"))
	t:SetText(name)
end

function mod:UPDATE_MOUSEOVER_UNIT()
	if not self:Active() or bit_band(db.behavior.showOn, 1) ~= 1 then return end
	
	self:CloseDisplay()
	if UnitIsPlayer("mouseover") and UnitIsFriend("player", "mouseover") then
		self:SendRequest(UnitName("mouseover"))
	end
	return true
end

function mod:PLAYER_TARGET_CHANGED()
	if not self:Active() or bit_band(db.behavior.showOn, 2) ~= 2 then return end
	self:CloseDisplay()
	if UnitIsPlayer("target") and UnitIsFriend("player", "target") then
		self:SendRequest(UnitName("target"))
	end
end

------------------------------------------------------------------------
------------------------------------------------------------------------


function mod:OnCommReceived( prefix, data, distribution, sender)
	if data == REQUEST then
		self:SendInfo(sender, distribution)
	else
		mod:ParseResponse(sender, strsplit("\1", data))
	end
end

function mod:ParseResponse(character, prefix, suffix, title, profile, flags)
	hive[character] = hive[character] or {}
	local t = hive[character]
	t.prefix = strtrim(prefix or "")
	t.suffix = strtrim(suffix or "")
	
	t.profile = strtrim(profile)
	t.profile = #t.profile == 0 and L["(No profile)"] or t.profile
	
	t.title = strtrim(title or "")
	t.flags = tonumber(strtrim(flags or "0"), 16)
	mod:Display(character, t)
end

function mod:SendRequest(to)
	local t = GetTime()
	local throttle = hive[ot] and CLIENT_THROTTLE_QUERY_TIME or THROTTLE_QUERY_TIME
	if t - (queryTimes[to] or 0) < throttle then
		if hive[to] then
			mod:Display(to)
		end
		return
	end
	queryTimes[to] = t

	self:SendCommMessage(PREFIX, REQUEST, "WHISPER", to)
end
	
do
	local info = {}
	function mod:SendInfo(to, distribution)
		wipe(info)
		tinsert(info, db.character.prefix)
		tinsert(info, db.character.suffix)
		tinsert(info, db.character.title)
		tinsert(info, db.character.profile)
		tinsert(info, ("%x"):format(bit_bor(db.role_mask, db.rp_level_mask)))
		local text = tconcat(info, "\1")
		self:SendCommMessage(PREFIX, text, distribution, distribution == "WHISPER" and to)
	end
end

function mod:Display(who, data)
	data = data or hive[who]
	display:ClearLines()
	display.name:SetText(getFullName(nil, who))
	display.title:SetText(data.title or "")
	display.profile:SetText(data.profile)
	display.status:SetText(getStatusString(data.flags))
	display:SetWidth(math.max(display.name:GetWidth(), display.title:GetWidth(), 300) + 45)
	display.profile:SetWidth(display:GetWidth() - 20)
	display:SetHeight(display.name:GetHeight() + display.title:GetHeight() + display.status:GetHeight() + display.profile:GetHeight() + 40)
	display:Show()
end

function mod:CloseDisplay()
	if db.sticky then return end
	display:Hide()
end

function mod:CreateDisplay()
	if display then return display end
	display = CreateFrame("Frame", "MarySueProfileFrame", UIParent)
	tinsert(UISpecialFrames, "MarySueProfileFrame")
	
	local f = display
	f:SetBackdrop({
		bgFile = [[Interface\DialogFrame\UI-DialogBox-Background]],
	    edgeFile = [[Interface\Tooltips\UI-Tooltip-Border]],
	    tile = true, tileSize = 16, edgeSize = 16,
	    insets = { left = 5, right = 5, top = 5, bottom = 5 }
	})
	f:SetWidth(150)
	f:SetHeight(150)
	f:SetPoint("LEFT")
	
	f:EnableMouse(true)
	f:SetMovable(true)
	f:SetClampedToScreen(true)	
	f:RegisterForDrag("LeftButton")
	f:SetScript("OnDragStart", function(frame)
		frame:StartMoving()
	end)
	f:SetScript("OnDragStop", function(frame)
		frame:StopMovingOrSizing()
	end)	
	
	f.name = f:CreateFontString(nil, nil, "GameFontNormalLarge")
	f.name:SetText("Name")
	f.name:SetPoint("TOPLEFT", display, "TOPLEFT", 10, -10)
	f.name:SetTextColor(0.6, 0.7, 1.0)
	
	f.title = f:CreateFontString(nil, nil, "GameFontNormal")
	f.title:SetText("Title")
	f.title:SetPoint("TOPLEFT", f.name, "BOTTOMLEFT", 0, -5)
	f.title:SetTextColor(0.4, 0.5, 1.0)

	f.status = f:CreateFontString(nil, nil, "GameFontNormal")
	f.status:SetText("Status")
	f.status:SetPoint("TOPLEFT", f.title, "BOTTOMLEFT", 0, -5)
	f.status:SetTextColor(1, 1, 0.8, 0.8)
	
	f.profile = f:CreateFontString(nil, nil, "GameFontNormal")
	f.profile:SetText("Profile")
	f.profile:SetJustifyV("TOP")
	f.profile:SetJustifyH("LEFT")
	f.profile:SetPoint("TOPLEFT", f.status, "BOTTOMLEFT", 0, -5)
	f.profile:SetTextColor(1, 1, 1)
	
	f.close = CreateFrame("Button", nil, f, "UIPanelCloseButton")
	f.close:SetPoint("TOPRIGHT")
	
	f.ClearLines = function(self)
		self.name:SetText("")
		self.title:SetText("")
		self.status:SetText("")
		self.profile:SetText("")
	end
	
	return display
end

function mod:UpdateWindow()
	if db.sticky then
		display.close:Show()
	else
		display.close:Hide()
	end
end

function mod:Active()
	if db.behavior.disableInInstances then		
		return IsInInstance() ~= 1
	end
	return true
	
end

mod:CreateDisplay()